#!/usr/bin/env python

# Copyright 2020, TUF contributors
# SPDX-License-Identifier: MIT OR Apache-2.0

"""
<Program Name>
  fast_server_exit.py

<Author>
  Martin Vrachev.

<Started>
  October 29, 2020.

<Copyright>
  See LICENSE-MIT OR LICENSE for licensing information.

<Purpose>
  Used for tests in tests/test_utils.py.
"""

import sys

sys.exit(0)
